%-- This code has been realized to illustrate neural network concepts during 
%-- the workshop organized by the "club EEA" which happened in March 2018 
%-- https://rnn2018lyon.sciencesconf.org/
%--
%-- This code has been strongly inspired by the Mooc of Andrew Ng on coursera
%-- https://www.coursera.org/learn/machine-learning/
%-- 
%-- The application treated here concerns the automatic recognition of 
%-- digits drawn by hand. This code is not protected by any license, you 
%-- can re-use it freely without any restriction. Please do not forget to
%-- cite this event for any use of the code.

function [Theta1,Theta2] = train(X, y, param)

    %-- Trains neural network classifiers and returns the corresponding
    %-- parameters stored in matrix Theta1 & Theta2
    
    %-- Check whether all parameters are given
    if (~exist('param','var'))
        param.maxIter = 50;                 %-- maximum number of iterations
        param.input_layer_size  = 900;      %-- number of input units  (30x30 units to in our case)
        param.hidden_layer_size = 25;       %-- number of hidden units
        param.output_layer_size = 10;       %-- number of output units (10 units to recognize digits from 0 to 9 in our case)
        param.lambda = 1;                   %-- regularization parameter
        param.gradient_descent_scheme = 2;  %-- 1: simple gradient descente / 2: Polack-Ribiere conjugate gradient
    end
    
    %-- Maximum number of iterations
    if ~isfield(param,'maxIter')
        param.maxIter = 100;                %-- 100 by default
    end    
    
    %-- Number of input units
    if ~isfield(param,'input_layer_size')
        param.input_layer_size = 900;       %-- 900 by default
    end
    
    %-- Number of hidden units
    if ~isfield(param,'hidden_layer_size')
        param.hidden_layer_size = 25;       %-- 25 by default
    end
    
    %-- Number of output units
    if ~isfield(param,'output_layer_size')
        param.output_layer_size = 10;       %-- 10 units to recognize digits from 0 to 9 in our case
    end
    
    %-- Regularization parameter
    if ~isfield(param,'lambda')
        param.lambda = 1;
    end
    
    %-- Gradient descente strategy
    if ~isfield(param,'gradient_descent_scheme')
        param.gradient_descent_scheme = 2;  %-- 1: simple gradient descente / 2: Polack-Ribiere conjugate gradient
    end
    

    %-- Initialization parameters
    initial_Theta1 = nn.randInitializeWeights(param.input_layer_size, param.hidden_layer_size);
    initial_Theta2 = nn.randInitializeWeights(param.hidden_layer_size, param.output_layer_size);
    
    
    %-- Unroll parameters
    initial_nn_params = [initial_Theta1(:) ; initial_Theta2(:)];
    
       
    %-- Create "short hand" for the cost function to be minimized
    costFunc = @(p) nn.costFunction(p, param.input_layer_size, ...
                                       param.hidden_layer_size, ...
                                       param.output_layer_size, ...
                                       X, y, param.lambda);
    
    
    %-- Perform gradient descent method
    switch param.gradient_descent_scheme
        case 1      %-- simple gradient descent scheme
            disp('Perform simple gradient descent scheme')
            options = struct('MaxIter', param.maxIter);  
            [nn_params] = nn.gradient_descent(costFunc, initial_nn_params, options);
        case 2      %-- Polack-Ribiere conjugate gradients
            disp('Perform Polack-Ribiere conjugate gradients')
            options = struct('MaxIter', param.maxIter);    
            [nn_params,~] = nn.fmincg(costFunc, initial_nn_params, options);
    end
    
    
    %-- Extract Theta1 and Theta2 variables back from nn_params
    Theta1 = reshape(nn_params(1:param.hidden_layer_size * (param.input_layer_size+1)), ...
                 param.hidden_layer_size, (param.input_layer_size+1));

    Theta2 = reshape(nn_params((1+(param.hidden_layer_size * (param.input_layer_size+1))):end), ...
                 param.output_layer_size, (param.hidden_layer_size+1));
       
end
